@@ -3,32 +3,36 @@
 Emulates the AMQ API for non-AMQ transports.
 """
 
+from __future__ import annotations
+
 import base64
 import socket
 import sys
 import warnings
-
 from array import array
 from collections import OrderedDict, defaultdict, namedtuple
 from itertools import count
 from multiprocessing.util import Finalize
 from queue import Empty
-from time import sleep, monotonic
+from time import monotonic, sleep
+from typing import TYPE_CHECKING
 
 from amqp.protocol import queue_declare_ok_t
 
-from kombu.exceptions import ResourceError, ChannelError
+from kombu.exceptions import ChannelError, ResourceError
 from kombu.log import get_logger
-from kombu.utils.encoding import str_to_bytes, bytes_to_str
+from kombu.transport import base
 from kombu.utils.div import emergency_dump_state
+from kombu.utils.encoding import bytes_to_str, str_to_bytes
 from kombu.utils.scheduling import FairCycle
 from kombu.utils.uuid import uuid
 
-from kombu.transport import base
-
 from .exchange import STANDARD_EXCHANGE_TYPES
 
-ARRAY_TYPE_H = 'H' if sys.version_info[0] == 3 else b'H'
+if TYPE_CHECKING:
+    from types import TracebackType
+
+ARRAY_TYPE_H = 'H'
 
 UNDELIVERABLE_FMT = """\
 Message could not be delivered: No queues bound to exchange {exchange!r} \
@@ -97,6 +101,7 @@ class BrokerState:
 
     #: The queue index is used to access directly (constant time)
     #: all the bindings of a certain queue.  It has the following structure::
+    #:
     #:     {
     #:         queue: {
     #:             (queue, exchange, routing_key),
@@ -154,6 +159,7 @@ class QoS:
     Only supports `prefetch_count` at this point.
 
     Arguments:
+    ---------
         channel (ChannelT): Connection channel.
         prefetch_count (int): Initial prefetch count (defaults to 0).
     """
@@ -178,6 +184,8 @@ class QoS:
         self.channel = channel
         self.prefetch_count = prefetch_count or 0
 
+        # Standard Python dictionaries do not support setting attributes
+        # on the object, hence the use of OrderedDict
         self._delivered = OrderedDict()
         self._delivered.restored = False
         self._dirty = set()
@@ -204,7 +212,8 @@ class QoS:
         bulk 'get message' calls are preferred to many individual 'get message'
         calls - like SQS.
 
-        Returns:
+        Returns
+        -------
             int: greater than zero.
         """
         pcount = self.prefetch_count
@@ -266,6 +275,7 @@ class QoS:
         """Restore all unacknowledged messages at shutdown/gc collect.
 
         Note:
+        ----
             Can only be called once for each instance, subsequent
             calls will be ignored.
         """
@@ -294,11 +304,12 @@ class QoS:
             state.restored = True
 
     def restore_visible(self, *args, **kwargs):
-        """Restore any pending unackwnowledged messages.
+        """Restore any pending unacknowledged messages.
 
         To be filled in for visibility_timeout style implementations.
 
         Note:
+        ----
             This is implementation optional, and currently only
             used by the Redis transport.
         """
@@ -348,6 +359,7 @@ class AbstractChannel:
     you'd usually want to implement in a virtual channel.
 
     Note:
+    ----
         Do not subclass directly, but rather inherit
         from :class:`Channel`.
     """
@@ -372,6 +384,7 @@ class AbstractChannel:
         """Delete `queue`.
 
         Note:
+        ----
             This just purges the queue, if you need to do more you can
             override this method.
         """
@@ -381,6 +394,7 @@ class AbstractChannel:
         """Create new queue.
 
         Note:
+        ----
             Your transport can override this method if it needs
             to do something whenever a new queue is declared.
         """
@@ -388,7 +402,8 @@ class AbstractChannel:
     def _has_queue(self, queue, **kwargs):
         """Verify that queue exists.
 
-        Returns:
+        Returns
+        -------
             bool: Should return :const:`True` if the queue exists
                 or :const:`False` otherwise.
         """
@@ -407,6 +422,7 @@ class Channel(AbstractChannel, base.StdChannel):
     """Virtual channel.
 
     Arguments:
+    ---------
         connection (ConnectionT): The transport instance this
             channel is part of.
     """
@@ -463,14 +479,7 @@ class Channel(AbstractChannel, base.StdChannel):
             typ: cls(self) for typ, cls in self.exchange_types.items()
         }
 
-        try:
-            self.channel_id = self.connection._avail_channel_ids.pop()
-        except IndexError:
-            raise ResourceError(
-                'No free channel ids, current={}, channel_max={}'.format(
-                    len(self.connection.channels),
-                    self.connection.channel_max), (20, 10),
-            )
+        self.channel_id = self._get_free_channel_id()
 
         topts = self.connection.client.transport_options
         for opt_name in self.from_transport_options:
@@ -675,6 +684,7 @@ class Channel(AbstractChannel, base.StdChannel):
         """Change QoS settings for this channel.
 
         Note:
+        ----
             Only `prefetch_count` is supported.
         """
         self.qos.prefetch_count = prefetch_count
@@ -697,9 +707,10 @@ class Channel(AbstractChannel, base.StdChannel):
     def _lookup(self, exchange, routing_key, default=None):
         """Find all queues matching `routing_key` for the given `exchange`.
 
-        Returns:
-            str: queue name -- must return the string `default`
-                if no queues matched.
+        Returns
+        -------
+            list[str]: queue names -- must return `[default]`
+                if default is set and no queues matched.
         """
         if default is None:
             default = self.deadletter_queue
@@ -728,7 +739,8 @@ class Channel(AbstractChannel, base.StdChannel):
         message = message.serializable()
         message['redelivered'] = True
         for queue in self._lookup(
-                delivery_info['exchange'], delivery_info['routing_key']):
+            delivery_info['exchange'],
+                delivery_info['routing_key']):
             self._put(queue, message)
 
     def _restore_at_beginning(self, message):
@@ -764,7 +776,8 @@ class Channel(AbstractChannel, base.StdChannel):
     def flow(self, active=True):
         """Enable/disable message flow.
 
-        Raises:
+        Raises
+        ------
             NotImplementedError: as flow
                 is not implemented by the base virtual implementation.
         """
@@ -789,12 +802,12 @@ class Channel(AbstractChannel, base.StdChannel):
         self.exchange_types = None
 
     def encode_body(self, body, encoding=None):
-        if encoding:
+        if encoding and encoding.lower() != 'utf-8':
             return self.codecs.get(encoding).encode(body), encoding
         return body, encoding
 
     def decode_body(self, body, encoding=None):
-        if encoding:
+        if encoding and encoding.lower() != 'utf-8':
             return self.codecs.get(encoding).decode(body)
         return body
 
@@ -805,7 +818,12 @@ class Channel(AbstractChannel, base.StdChannel):
     def __enter__(self):
         return self
 
-    def __exit__(self, *exc_info):
+    def __exit__(
+        self,
+        exc_type: type[BaseException] | None,
+        exc_val: BaseException | None,
+        exc_tb: TracebackType | None
+    ) -> None:
         self.close()
 
     @property
@@ -832,6 +850,7 @@ class Channel(AbstractChannel, base.StdChannel):
         The value is limited to within a boundary of 0 to 9.
 
         Note:
+        ----
             Higher value has more priority.
         """
         try:
@@ -845,6 +864,22 @@ class Channel(AbstractChannel, base.StdChannel):
 
         return (self.max_priority - priority) if reverse else priority
 
+    def _get_free_channel_id(self):
+        # Cast to a set for fast lookups, and keep stored as an array
+        # for lower memory usage.
+        used_channel_ids = set(self.connection._used_channel_ids)
+
+        for channel_id in range(1, self.connection.channel_max + 1):
+            if channel_id not in used_channel_ids:
+                self.connection._used_channel_ids.append(channel_id)
+                return channel_id
+
+        raise ResourceError(
+            'No free channel ids, current={}, channel_max={}'.format(
+                len(self.connection.channels),
+                self.connection.channel_max), (20, 10),
+        )
+
 
 class Management(base.Management):
     """Base class for the AMQP management API."""
@@ -865,6 +900,7 @@ class Transport(base.Transport):
     """Virtual transport.
 
     Arguments:
+    ---------
         client (kombu.Connection): The client this is a transport for.
     """
 
@@ -872,9 +908,6 @@ class Transport(base.Transport):
     Cycle = FairCycle
     Management = Management
 
-    #: Global :class:`BrokerState` containing declared exchanges and bindings.
-    state = BrokerState()
-
     #: :class:`~kombu.utils.scheduling.FairCycle` instance
     #: used to fairly drain events from channels (set by constructor).
     cycle = None
@@ -902,6 +935,8 @@ class Transport(base.Transport):
 
     def __init__(self, client, **kwargs):
         self.client = client
+        # :class:`BrokerState` containing declared exchanges and bindings.
+        self.state = BrokerState()
         self.channels = []
         self._avail_channels = []
         self._callbacks = {}
@@ -909,9 +944,7 @@ class Transport(base.Transport):
         polling_interval = client.transport_options.get('polling_interval')
         if polling_interval is not None:
             self.polling_interval = polling_interval
-        self._avail_channel_ids = array(
-            ARRAY_TYPE_H, range(self.channel_max, 0, -1),
-        )
+        self._used_channel_ids = array(ARRAY_TYPE_H)
 
     def create_channel(self, connection):
         try:
@@ -923,7 +956,11 @@ class Transport(base.Transport):
 
     def close_channel(self, channel):
         try:
-            self._avail_channel_ids.append(channel.channel_id)
+            try:
+                self._used_channel_ids.remove(channel.channel_id)
+            except ValueError:
+                # channel id already removed
+                pass
             try:
                 self.channels.remove(channel)
             except ValueError:
@@ -936,7 +973,7 @@ class Transport(base.Transport):
         # this channel is then used as the next requested channel.
         # (returned by ``create_channel``).
         self._avail_channels.append(self.create_channel(self))
-        return self     # for drain events
+        return self  # for drain events
 
     def close_connection(self, connection):
         self.cycle.close()
